#include <iostream>

#include <Windows.h>

#include <Varjo_plugin.h>

#include "PluginStatusCodes.hpp"
#include "ExampleTracker.hpp"

/**
 * Provide plugin types supported by the plugin library
 */
VARJO_API varjo_PluginType varjo_GetPluginType()
{  //
    std::cout << __FUNCTION__ << "(): " << std::endl;
    return varjo_PluginType_HMD_Tracker;
}

/**
 * Provide the API version number supported by the plugin.
 * Use @ref VARJO_PLUGIN_API_VERSION macro to indicate supported API version number.
 */
VARJO_API uint64_t varjo_GetPluginAPIVersion()
{  //
    std::cout << __FUNCTION__ << "(): " << std::endl;
    return VARJO_PLUGIN_API_VERSION;
}

/**
 * @returns Null-terminated C-string containing plugin name
 */
VARJO_API const char* varjo_GetPluginName()
{  //
    std::cout << __FUNCTION__ << "(): " << std::endl;
    return "example plugin";
}

/**
 * Return a string description of a given status code
 */
VARJO_API const char* varjo_GetPluginStatusDescription(varjo_PluginStatus pluginStatus)
{  //
    std::cout << __FUNCTION__ << "(): " << std::endl;

    switch (pluginStatus) {
        case varjo_PluginStatus_OK: {
            return "No error";
        }
        case varjo_PluginStatus_NullPointer: {
            return "Function called with a null pointer";
        }
        case varjo_PluginStatus_TrackerInitializationError: {
            return "Couldn't initialize tracker";
        }
        case varjo_PluginStatus_TrackingNotInitialized: {
            return "Tracker is not intialized";
        }
        default: {
            break;
        }
    }

    return "Unknown error";
}

/**
 * Initialize plugin
 */
VARJO_API varjo_PluginStatus varjo_PluginInit(        //
    const varjo_PluginUtilityAPI* pluginUtilityAPI,   //
    varjo_PluginUtilitySession pluginUtilitySession,  //
    varjo_PluginContext* context)
{
    std::cout << __FUNCTION__ << "(): " << std::endl;
    if (!pluginUtilityAPI) {
        return varjo_PluginStatus_NullPointer;
    }

    ExampleTracker* tracker = new ExampleTracker(  //
        pluginUtilityAPI, pluginUtilitySession);
    *context = reinterpret_cast<varjo_PluginContext>(tracker);

    return varjo_PluginStatus_OK;
}

/**
 * Deinitializes plugin
 */
VARJO_API varjo_PluginStatus varjo_PluginShutdown(varjo_PluginContext* context)
{
    std::cout << __FUNCTION__ << "(): " << std::endl;
    ExampleTracker* tracker = reinterpret_cast<ExampleTracker*>(*context);
    if (!tracker) {
        return varjo_PluginStatus_NullPointer;
    }

    delete tracker;
    *context = nullptr;

    return varjo_PluginStatus_OK;
}

/**
 * Start tracker processing
 */
VARJO_API varjo_PluginStatus varjo_TrackerStart(varjo_PluginContext context)
{
    std::cout << __FUNCTION__ << "(): " << std::endl;
    ExampleTracker* tracker = reinterpret_cast<ExampleTracker*>(context);
    if (!tracker) {
        return varjo_PluginStatus_NullPointer;
    }

    return tracker->start();
}

/**
 * Stop tracker processing
 */
VARJO_API varjo_PluginStatus varjo_TrackerStop(varjo_PluginContext context)
{
    std::cout << __FUNCTION__ << "(): " << std::endl;
    ExampleTracker* tracker = reinterpret_cast<ExampleTracker*>(context);
    if (!tracker) {
        return varjo_PluginStatus_NullPointer;
    }

    return tracker->stop();
}

/**
 * Return the tracked HMD pose for a given time
 */
VARJO_API varjo_PluginStatus varjo_GetHMDPose(  //
    varjo_PluginContext context,                //
    varjo_Nanoseconds timeStamp,                //
    varjo_DevicePose* pose)
{
    ExampleTracker* tracker = reinterpret_cast<ExampleTracker*>(context);
    if (!tracker || !pose) {
        return varjo_PluginStatus_NullPointer;
    }

    tracker->getHMDPose(timeStamp, pose);

    return varjo_PluginStatus_OK;
}

/**
 * Plugin DLL entry point
 */
BOOL WINAPI DllMain(  //
    IN HINSTANCE hDllHandle, IN DWORD nReason, IN LPVOID Reserved)
{
    switch (nReason) {
        case DLL_PROCESS_ATTACH:  //
            break;

        case DLL_PROCESS_DETACH:  //
            break;
    }

    return TRUE;
}
