// Copyright 2019-2021 Varjo Technologies Oy. All rights reserved.

#ifndef VARJO_TYPES_H
#define VARJO_TYPES_H

#include <cstdint>

#if defined __cplusplus
extern "C" {
#endif

#if defined(_WIN32)
#define VARJO_FNTABLE_CALLTYPE __cdecl
#else
#define VARJO_FNTABLE_CALLTYPE
#endif

/**
 * Type representing logical values - True and False.
 */
typedef int32_t varjo_Bool;
static const varjo_Bool varjo_False = 0;  //!< Value representing False
static const varjo_Bool varjo_True = 1;   //!< Value representing True

/**
 * Time in nanoseconds. Absolute times are relative to an epoch which is constant during execution of the program.
 */
typedef int64_t varjo_Nanoseconds;

/**
 * 64bit floating point 3D vector.
 */
typedef struct varjo_Vector3D {
    double x;  //!< X coordinate.
    double y;  //!< Y coordinate.
    double z;  //!< Z coordinate.
} varjo_Vector3D;

/**
 * Double precision quaternion.
 */
typedef struct varjo_Quaternion {
    double w;  //!< Quaternion component w
    double x;  //!< Quaternion component x
    double y;  //!< Quaternion component y
    double z;  //!< Quaternion component z
} varjo_Quaternion;


/**
 * Double precision 3x3 matrix.
 * The matrix usage convention is that values are stored in column-major order.
 */
typedef struct varjo_Matrix3x3 {
    double value[9];
} varjo_Matrix3x3;

/**
 * Double precision 4x4 matrix.
 * The matrix usage convention is that values are stored in column-major order.
 */
typedef struct varjo_Matrix4x4 {
    double value[16];
} varjo_Matrix4x4;

/**
 * Bit-field type determining tracking status
 */
typedef uint64_t varjo_PoseFlags;
static const varjo_PoseFlags varjo_PoseFlags_Ok = 0x1;                    //!< Flag indicating that the tracking is OK
static const varjo_PoseFlags varjo_PoseFlags_TrackingLost = 0x2;          //!< Flag indicating that no tracking data is available
static const varjo_PoseFlags varjo_PoseFlags_TrackingDisconnected = 0x4;  //!< Flag indicating that the tracker is disconnected
static const varjo_PoseFlags varjo_PoseFlags_HasPosition = 0x08;          //!< Flag indicating that the pose has position information
static const varjo_PoseFlags varjo_PoseFlags_HasRotation = 0x10;          //!< Flag indicating that the pose has rotation information
static const varjo_PoseFlags varjo_PoseFlags_HasVelocity = 0x20;          //!< Flag indicating that the pose has velocity information
static const varjo_PoseFlags varjo_PoseFlags_HasAngularVelocity = 0x40;   //!< Flag indicating that the pose has angular velocity information
static const varjo_PoseFlags varjo_PoseFlags_HasAcceleration = 0x80;      //!< Flag indicating that the pose has acceleration information
static const varjo_PoseFlags varjo_PoseFlags_HasConfidence = 0x100;       //!< Flag indicating that the pose has confidence information

/**
 * Structure describing tracked device pose.
 *
 * <em>
 * Coordinate system is right-handed, positive X goes right, positive Y goes up
 * and negative Z goes forward from HMD users perspective.
 * Distance unit is meters.
 * </em>
 */
typedef struct varjo_DevicePose {
    varjo_Nanoseconds timeStamp;     //!< Pose timestamp (ns)
    int64_t frameNo;                 //!< Frame number, increases monotonically.
    varjo_Vector3D position;         //!< Device position (m)
    varjo_Quaternion rotation;       //!< Device rotation
    varjo_Vector3D velocity;         //!< Device linear velocity (m/s)
    varjo_Vector3D angularVelocity;  //!< Device angular velocity (radians/s)
    varjo_Vector3D acceleration;     //!< Device acceleration (m/s^2)
    varjo_PoseFlags poseFlags;       //!< Bit field value describing pose
    double confidence;               //!< Tracker confidence
} varjo_DevicePose;


/**
 * Bit-field type determining capabilities of IMU hardware
 */
typedef uint64_t varjo_IMUCapability;
static const varjo_IMUCapability varjo_IMUCapability_Acceleration = 0x01;     //!< IMU HW is capable of reporting accelerometer measurements
static const varjo_IMUCapability varjo_IMUCapability_AngularVelocity = 0x02;  //!< IMU HW is capable of reporting gyro measurements
static const varjo_IMUCapability varjo_IMUCapability_Attitude = 0x04;         //!< IMU HW is capable of reporting magnetometer measurements

/**
 * Bit-field type determining which values of IMU data are off-scale.
 * These flags are set when a corresponding sensor reading exceeds a limit of it's
 * measurement range
 */
typedef uint64_t varjo_IMUDataOffscale;
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AccelerationX = 0x001;     //!< Flag indicating the X-axis acceleration measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AccelerationY = 0x002;     //!< Flag indicating the Y-axis acceleration measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AccelerationZ = 0x004;     //!< Flag indicating the Z-axis acceleration measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AngularVelocityX = 0x008;  //!< Flag indicating the X-axis angular velocity measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AngularVelocityY = 0x010;  //!< Flag indicating the Y-axis angular velocity measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AngularVelocityZ = 0x020;  //!< Flag indicating the Z-axis angular velocity measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AttitudeX = 0x040;         //!< Flag indicating the X-axis attitude measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AttitudeY = 0x080;         //!< Flag indicating the Y-axis attitude measurement is off-scale
static const varjo_IMUDataOffscale varjo_IMUDataOffscale_AttitudeZ = 0x100;         //!< Flag indicating the Z-axis attitudey measurement is off-scale

/**
 * Bit-field type determining which values of IMU data are not valid,
 * e.g. in case the IMU HW can not provide corresponding measurement type, or
 * does not have a measurement sample
 */
typedef uint64_t varjo_IMUDataInvalid;
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AccelerationX = 0x001;     //!< Flag indicating the X-axis acceleration measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AccelerationY = 0x002;     //!< Flag indicating the Y-axis acceleration measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AccelerationZ = 0x004;     //!< Flag indicating the Z-axis acceleration measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AngularVelocityX = 0x008;  //!< Flag indicating the X-axis angular velocity measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AngularVelocityY = 0x010;  //!< Flag indicating the Y-axis angular velocity measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AngularVelocityZ = 0x020;  //!< Flag indicating the Z-axis angular velocity measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AttitudeX = 0x040;         //!< Flag indicating the X-axis attitude measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AttitudeY = 0x080;         //!< Flag indicating the Y-axis attitude measurement is not valid
static const varjo_IMUDataInvalid varjo_IMUDataInvalid_AttitudeZ = 0x100;         //!< Flag indicating the Z-axis attitude measurement is not valid


/**
 * Type identifying IMU component type
 */
typedef int64_t varjo_IMUComponent;
static const varjo_IMUComponent varjo_IMUComponent_Accelerometer = 1;  //!< Constant identifying IMU component accelerometer
static const varjo_IMUComponent varjo_IMUComponent_Gyro = 2;           //!< Constant identifying IMU component gyro
static const varjo_IMUComponent varjo_IMUComponent_Magnetometer = 3;   //!< Constant identifying IMU component magnetometer

/**
 * Structure containing calibration parameters for an IMU device
 */
typedef struct varjo_IMUComponentCalibrationInfo {
    varjo_Matrix3x3 scale;  //!< Calibration scale
    varjo_Vector3D bias;    //!< Calibration bias
} varjo_IMUDeviceCalibrationInfo;

/**
 * Structure describing a sample of IMU data
 */
typedef struct varjo_ImuSample {
    varjo_Nanoseconds timeStamp;          //!< IMU sample timestamp (ns)
    varjo_Vector3D acceleration;          //!< Device acceleration (m/s^2)
    varjo_Vector3D angularVelocity;       //!< Device angular velocity (radians/s)
    varjo_Vector3D attitude;              //!< Device attitude
    varjo_IMUDataOffscale offscaleFlags;  //!< Offscale flags
    varjo_IMUDataInvalid validityFlags;   //!< Validity flags
} varjo_ImuSample;

/**
 * Bit-field type determining type of plugin API
 */
typedef uint64_t varjo_PluginType;
static const varjo_PluginType varjo_PluginType_HMD_Tracker = 0x01;     //!< Flag indicating this plugin provides HMD tracking functionality
static const varjo_PluginType varjo_PluginType_MotionPlatform = 0x02;  //!< Flag indicating this plugin provides motion platform functionality

/**
 * Plugin status
 */
typedef int64_t varjo_PluginStatus;
static const varjo_PluginStatus varjo_PluginStatus_OK = 0;  //!< Plugin status OK

/**
 * Type determining type of tracked device
 */
typedef int64_t varjo_DeviceType;
static const varjo_PluginType varjo_DeviceType_HMD = 1;  //!< Device type - Head Mounted Display

/**
 * Plugin session context
 */
typedef struct varjo_PluginContext_t* varjo_PluginContext;

/**
 * @brief varjo_PluginUtilitySession
 */
typedef struct varjo_PluginUtilitySession_t* varjo_PluginUtilitySession;

/**
 * Settings key enumeration handle
 */
typedef struct varjo_ConfigurationKeyIterator_t* varjo_ConfigurationKeyIterator;

/**
 * Plugin system state
 */
typedef int64_t varjo_PluginSystemState;
static const varjo_PluginSystemState varjo_PluginSystemState_OK = 0;       //!< Plugin system is functioning correctly
static const varjo_PluginSystemState varjo_PluginSystemState_Warning = 1;  //!< Plugin system is reporting a state requiring user attention e.g. battery low
static const varjo_PluginSystemState varjo_PluginSystemState_CriticalError = 2;  //!< Plugin system is reporting a critical error state, e.g. tracking failure

/**
 * @struct varjo_PluginUtilityAPI
 *
 * Utility functions plugins can call
 */
typedef struct varjo_PluginUtilityAPI {
    /**
     * Returns Varjo product id string
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @returns NULL terminated C-string containing Varjo product id. e.g. "VR-1"
     */
    typedef const char*(VARJO_FNTABLE_CALLTYPE* GetProductId_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetProductId_t. */
    GetProductId_t GetProductId;

    /**
     * Returns hardware capabilities of the IMU
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @returns A combination of @ref varjo_IMUCapability flags describing IMU hardware
     * capabilities.
     */
    typedef varjo_IMUCapability(VARJO_FNTABLE_CALLTYPE* GetIMUCapabilities_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetIMUCapabilities_t. */
    GetIMUCapabilities_t GetIMUCapabilities;

    /**
     * Returns transformation matrix between IMU and the HMD coordinate systems
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @returns A transformation matrix between IMU and HMD coordinate systems.
     */
    typedef varjo_Matrix4x4(VARJO_FNTABLE_CALLTYPE* GetIMUToHMDTransform_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetIMUToHMDTransform_t. */
    GetIMUToHMDTransform_t GetIMUToHMDTransform;

    /**
     * Returns calibration for an IMU component
     *
     * <em>
     * Calibration parameters are calculated after production and are constant in run time.
     *
     * Calibration can be applied this way: <c>IMUCCalibrated = (IMURaw - Bias) * Scale</c>
     * </em>
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param component - IMU hardare component for which the calibration is requested
     * @returns IMU calibration parameters for a @c component
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* GetIMUComponentCalibration_t)(  //
        varjo_PluginUtilitySession session,                              //
        varjo_IMUComponent component,                                    //
        varjo_IMUComponentCalibrationInfo* calibrationInfo);

    /** @sa GetIMUComponentCalibration_t. */
    GetIMUComponentCalibration_t GetIMUComponentCalibration;

    /**
     * Reads up to @c sampleCount samples of @ref varjo_ImuSample into a @c samples buffer.
     *
     * The @c varjo_ImuSample values obtained via @ref ReadIMUSamples_t are not calibrated.
     * @sa GetIMUComponentCalibration_t
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param samples - Pointer to the buffer where the IMU data samples should be stored
     * @param sampleCount - Number of IMU samples to read
     * @returns Number of IMU samples read
     */
    typedef int32_t(VARJO_FNTABLE_CALLTYPE* ReadIMUSamples_t)(  //
        varjo_PluginUtilitySession session,                     //
        varjo_ImuSample* samples, int32_t sampleCount);

    /** @sa ReadIMUSamples_t. */
    ReadIMUSamples_t ReadIMUSamples;

    /**
     * Atomically applies set configuration
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     */
    typedef varjo_Bool(VARJO_FNTABLE_CALLTYPE* SyncConfiguration_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa SyncConfiguration_t. */
    SyncConfiguration_t SyncConfiguration;

    /**
     * Enumerates configuration keys
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param handle - enumeration handle, pass NULL to start enumeration
     * @return NULL when enumeration is over, otherwise a null-terminated C-string containing
     *         key name.
     */
    typedef const char*(VARJO_FNTABLE_CALLTYPE* EnumConfigurationKeys_t)(  //
        varjo_PluginUtilitySession session,                                //
        varjo_ConfigurationKeyIterator* handle);

    /** @sa EnumConfigurationKeys_t. */
    EnumConfigurationKeys_t EnumConfigurationKeys;

    /**
     * Sets configuration entry @param value for @param key
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* SetConfigurationString_t)(  //
        varjo_PluginUtilitySession session,                          //
        const char* key, const char* value);

    /** @sa SetConfigurationString_t. */
    SetConfigurationString_t SetConfigurationString;

    /**
     * Writes configuration entry @param value for @param key
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @return If @c value is NULL, returns size of buffer necessary for @c value,
     * otherwise length of the value for @c key written to @c value.
     * Returs 0 if the requested configuration key is not found.
     */
    typedef int32_t(VARJO_FNTABLE_CALLTYPE* GetConfigurationString_t)(  //
        varjo_PluginUtilitySession session,                             //
        const char* key, char* value, int32_t valueLength);

    /** @sa GetConfigurationString_t. */
    GetConfigurationString_t GetConfigurationString;


    /**
     * Returns current system time
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @return Current time
     */
    typedef varjo_Nanoseconds(VARJO_FNTABLE_CALLTYPE* GetCurrentSystemTime_t)(  //
        varjo_PluginUtilitySession session);

    /** @sa GetCurrentSystemTime_t. */
    GetCurrentSystemTime_t GetCurrentSystemTime;

    /**
     * Writes a @c message to the log
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* WriteLogMessage_t)(  //
        varjo_PluginUtilitySession session,                   //
        const char* message);

    /** @sa WriteLogMessage_t. */
    WriteLogMessage_t WriteLogMessage;

    /**
     * Indicates a plugin system @c state to Varjo SW stack.
     *
     * @param session - Plugin utility session context acquired from @ref varjo_PluginInit
     * @param state - a value of type @ref varjo_PluginSystemState indicating plugin system state
     * @param message - optional message containing technical description of given state
     */
    typedef void(VARJO_FNTABLE_CALLTYPE* SetPluginSystemState_t)(  //
        varjo_PluginUtilitySession session,                        //
        varjo_PluginSystemState state,                             //
        const char* message);

    /** @sa SetPluginSystemState_t. */
    SetPluginSystemState_t SetPluginSystemState;

} varjo_PluginUtilityAPI;

#if defined __cplusplus
}
#endif

#endif  // VARJO_TYPES_H
